// Rabbits and Wolves model of predator-prey dynamics.
// Author: Aaron Weeden, Shodor, 2016

/*******************************************************************************
 * IMPORT LIBRARIES
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*******************************************************************************
 * DEFINE GLOBAL CONSTANTS
 ******************************************************************************/
#define POPULATION_COUNT 2 // The number of populations (rabbits and wolves)
#define TIME_STEP_COUNT 5000 // The simulation runs for this many steps
#define CELL_COUNT_Y 50 // The world has this many rows
#define CELL_COUNT_X 50 // The world has this many columns
#define MAX_GRASS 10 // A cell cannot have more grass than this
#define RABBIT_INIT_COUNT 40 // At the beginning, there are this many rabbits
#define WOLF_INIT_COUNT 10 // At the beginning, there are this many wolves
#define RABBIT_CARRYING_CAPACITY 120 // There cannot be more rabbits than this
#define WOLF_CARRYING_CAPACITY 20 // There cannot be more wolves than this
#define RABBIT_REPROD_CHANCE 40 /* A rabbit has this % likelihood of
                                    reproducing */
#define WOLF_REPROD_CHANCE 20 // A wolf has this % likelihood of reproducing
#define RABBIT_MEAL_SIZE 5 // A rabbit eats likes to eat this much in one
                           // sitting
#define WOLF_MEAL_SIZE 40 // A wolf likes to eat this much in one sitting
#define RABBIT_INIT_FOOD 3 /* A rabbit starts out with this much food in its
                              belly */
#define WOLF_INIT_FOOD 20 // A wolf starts out with this much food in its belly
#define RABBIT_MAX_FOOD 20 // A rabbit cannot have more food than this in its
                           // belly
#define WOLF_MAX_FOOD 80 // A wolf cannot have more food than this in its belly
#define RABBIT_HUNGER 5 // Rabbits lose this much food each time step
#define WOLF_HUNGER 1 // Wolves lose this much food each time step
#define RANDOM_SEED 0 // The random number generator is seeded with this value

/*******************************************************************************
 * DEFINE CUSTOM DATATYPES
 ******************************************************************************/
// Define an enumerated type for distinguishing agent types
typedef enum { NONE, RABBIT, WOLF } AgentType_t;

// Declare custom datatypes
struct Agent_t;
struct Cell_t;
struct Population_t;

// Define custom datatype for agent
typedef struct Agent_t {
  AgentType_t type; // What type of agent is it
  int index; // What is the agent's index within its population
  struct Cell_t * cell_p; // Which cell is the agent in
  int food; // How much food does the agent have in its belly
  struct Population_t * pop_p; // Which population does this agent belong to
} Agent_t;

// Define custom datatype for world cell
typedef struct Cell_t {
  Agent_t * agent_p; // What agent is in the cell
  int grass; // How much grass is in the cell
  int neighborCount; // How many neighboring cells does this cell have
  struct Cell_t * neighbor_ps[4]; // What are this cell's neighbors
} Cell_t;

// Define custom datatype for population of agents
typedef struct Population_t {
  AgentType_t type; // This is the type of agents that make up this population
  Agent_t * agents; // This is an array of the agents
  int count; // This counts how many agents there are
  int capacity; // The world cannot have more agents than this
  int reprodChance; /* If an agent is next to an empty cell, it will reproduce
                       with this percent chance */
  int mealSize; // Agents eat no more than this much food in one sitting
  int initFood; // Each rabbit starts with this much food in its belly
  int maxFood; // An agent cannot have more food than this in its belly
  int hunger; // An agent loses this much food each time step
  AgentType_t preyType; /* If this is a population of predators,
                           this is the type of its prey */
  void (*eat)(Agent_t * const agent_p); /* This is the function pointer to use
                                           when it is time to call an eat
                                           function */
} Population_t;

/*******************************************************************************
 * DEFINE GLOBAL VARIABLES
 ******************************************************************************/
// Define global array of populations
Population_t Populations[POPULATION_COUNT] = {
  {
    .type = RABBIT,
    .count = RABBIT_INIT_COUNT,
    .capacity = RABBIT_CARRYING_CAPACITY,
    .reprodChance = RABBIT_REPROD_CHANCE,
    .mealSize = RABBIT_MEAL_SIZE,
    .initFood = RABBIT_INIT_FOOD,
    .maxFood = RABBIT_MAX_FOOD,
    .hunger = RABBIT_HUNGER,
    .preyType = NONE
  },
  {
    .type = WOLF,
    .count = WOLF_INIT_COUNT,
    .capacity = WOLF_CARRYING_CAPACITY,
    .reprodChance = WOLF_REPROD_CHANCE,
    .mealSize = WOLF_MEAL_SIZE,
    .initFood = WOLF_INIT_FOOD,
    .maxFood = WOLF_MAX_FOOD,
    .hunger = WOLF_HUNGER,
    .preyType = RABBIT
  }
};

// Declare global array of cells
Cell_t Cells[CELL_COUNT_Y][CELL_COUNT_X];

/*******************************************************************************
 * DECLARE FUNCTIONS
 ******************************************************************************/
void Initialize();
  void AllocateMemory();
  void InitializeCells();
  void InitializeAgents();
    int GetRandom(int const max);
    void AddAgent(Population_t * const pop_p, Cell_t * const cell_p,
                  int const food);
void Simulate();
  void AllHungryAgentsStarve();
    void RemoveAgent(Agent_t * const agent_p);
  void AllAgentsGetHungry();
  void AllAgentsTryToEat();
  void APreyTriesToEat(Agent_t * const agent_p);
    int GetMin(int const a, int const b, int const c);
  void APredatorTriesToEat(Agent_t * const agent_p);
    Cell_t * GetRandomNeighborCell(Cell_t const * const cell_p,
                                   AgentType_t const type);
  void AllAgentsTryToReproduce();
  void AllAgentsTryToMove();
  void GrassGrows();
void Finalize();
  void FreeMemory();

/*******************************************************************************
 * DEFINE FUNCTIONS
 ******************************************************************************/
// Define the main function
int main() {
  Initialize();
  Simulate();
  Finalize();
  return 0;
}

// Define a function to initialize the model
void Initialize() {
  srandom(RANDOM_SEED);
  AllocateMemory();
  InitializeCells();
  InitializeAgents();
}

// Define a function which allocates enough space for the arrays of each
// population
void AllocateMemory() {
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Allocate memory
    pop_p->agents = (Agent_t *)malloc(pop_p->capacity * sizeof(Agent_t));
  }
}

// Define a function to initialize all the world cells - give them the full
// amount of grass and point to their neighbors
void InitializeCells() {
  // For each cell
  for (int y = 0; y < CELL_COUNT_Y; y++) {
    for (int x = 0; x < CELL_COUNT_X; x++) {
      // Create a temporary variable to help readability
      Cell_t * const cell_p = &(Cells[y][x]);

      // Set the full amount of grass on the cell
      cell_p->grass = MAX_GRASS;

      // Point to the cell's neighbors
      cell_p->neighborCount = 0;

      // Point to the top neighbor, if there is one
      if (y > 0) {
        cell_p->neighbor_ps[(cell_p->neighborCount)++] = &(Cells[y-1][x]);
      }

      // Point to the bottom neighbor, if there is one
      if (y < CELL_COUNT_Y - 1) {
        cell_p->neighbor_ps[(cell_p->neighborCount)++] = &(Cells[y+1][x]);
      }

      // Point to the left neighbor, if there is one
      if (x > 0) {
        cell_p->neighbor_ps[(cell_p->neighborCount)++] = &(Cells[y][x-1]);
      }

      // Point to the right neighbor, if there is one
      if (x < CELL_COUNT_X - 1) {
        cell_p->neighbor_ps[(cell_p->neighborCount)++] = &(Cells[y][x+1]);
      }
    }
  }
}

// Define a function to initialize all agents - give them a random (x,y)
// cell, point that cell at them, and set the agent's type and initial food
// level
void InitializeAgents() {
  // Make a list of open cells in the world
  Cell_t * openCell_ps[CELL_COUNT_Y * CELL_COUNT_X];
  int openCellCount = 0;
  for (int y = 0; y < CELL_COUNT_Y; y++) {
    for (int x = 0; x < CELL_COUNT_X; x++) {
      // Add the cell to the list
      openCell_ps[openCellCount++] = &(Cells[y][x]);
    }
  }

  // Give each agent a random cell, point that cell to the agent, remove
  // the cell from the list of open cells, and set the type and initial
  // food amount of the agent
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Decide which type of eat function the population should use.
    // If the population has no prey, then it must be a prey itself. Otherwise
    // the population is a predator
    pop_p->eat = pop_p->preyType == NONE ? APreyTriesToEat :
                                           APredatorTriesToEat;

    // Make a backup of the count variable before overwriting it
    int const count = pop_p->count;

    // Loop through each initial agent, giving it a random empty cell
    pop_p->count = 0;
    for (int agentIdx = 0; agentIdx < count; agentIdx++) {
      // Get a random empty cell for the agent
      int r = GetRandom(openCellCount);

      // Create the agent
      AddAgent(pop_p, openCell_ps[r], pop_p->initFood);

      // Remove the cell from the list of open cells
      for (int cellIdx = r; cellIdx < openCellCount - 1; cellIdx++) {
        openCell_ps[cellIdx] = openCell_ps[cellIdx + 1];
      }
      openCellCount--;
    }
  }
}

// Define a function that returns a random whole number
int GetRandom(int const max) {
  return max * random() / RAND_MAX;
}

// Create an agent and add it to the given population and cell with the given
// amount of initial food
void AddAgent(Population_t * const pop_p, Cell_t * const cell_p,
              int const food) {
  // Create a temporary variable to help readability
  Agent_t * const agent_p = &(pop_p->agents[pop_p->count]);

  // Assign the agent its initial properties
  agent_p->type = pop_p->type;
  agent_p->index = pop_p->count;
  agent_p->cell_p = cell_p;
  agent_p->food = food;
  agent_p->pop_p = pop_p;

  // Update the population counter
  pop_p->count++;

  // Point the cell to the agent
  cell_p->agent_p = agent_p;
}

// Define a function that runs the model for each time step
void Simulate() {
  for (int timeIdx = 0; timeIdx < TIME_STEP_COUNT; timeIdx++) {
    AllHungryAgentsStarve();
    AllAgentsGetHungry();
    AllAgentsTryToEat();
    AllAgentsTryToReproduce();
    AllAgentsTryToMove();
    GrassGrows();
  }
}

// Define a function that removes agents from the model whose food level is 0
void AllHungryAgentsStarve() {
  // Loop over each population
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Loop over each agent in the population
    for (int agentIdx = 0; agentIdx < pop_p->count; agentIdx++) {
      // Create a temporary variable to help readability
      Agent_t * const agent_p = &(pop_p->agents[agentIdx]);

      // If the agent has no food, remove the agent and update the loop index
      if (agent_p->food <= 0) {
        // Remove the agent
        RemoveAgent(agent_p);

        // Update the loop index
        agentIdx--;
      }
    }
  }
}

// Define a function that removes an agent from the model
void RemoveAgent(Agent_t * const agent_p) {
  // Create temporary variables to help readability
  Population_t * const pop_p = agent_p->pop_p;
  Agent_t * const agents = pop_p->agents;
  int * const count_p = &(pop_p->count);

  // Point the agent's cell at nothing
  agent_p->cell_p->agent_p = NULL;

  // Remove the agent from the array by shifting everything in the array to
  // the left
  for (int agentIdx = agent_p->index; agentIdx < (*count_p) - 1; agentIdx++) {
    // Create a temporary variable to help readability
    Agent_t * const thisAgent_p = &(agents[agentIdx]);

    // Replace the agent in the array with the agent to its right in the array
    (*thisAgent_p) = agents[agentIdx + 1];

    // Make sure the new agent's cell still points to it
    thisAgent_p->cell_p->agent_p = thisAgent_p;

    // Update the agent's index
    thisAgent_p->index = agentIdx;
  }

  // Update the population counter
  (*count_p)--;
}

// Define a function that reduces the food amount for each agent
void AllAgentsGetHungry() {
  // Loop over each population
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Loop over each agent
    for (int agentIdx = 0; agentIdx < pop_p->count; agentIdx++) {
      pop_p->agents[agentIdx].food -= pop_p->hunger;
    }
  }
}

// Define a function that has prey eat grass and predators eat prey
void AllAgentsTryToEat() {
  // Loop over all populations
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Loop over all agents in the population
    for (int agentIdx = 0; agentIdx < pop_p->count; agentIdx++) {
      // Call the population's eat function for the agent
      pop_p->eat(&(pop_p->agents[agentIdx]));
    }
  }
}

// Define a function that has each rabbit in the given population try to eat
// some grass
void APreyTriesToEat(Agent_t * const agent_p) {
  // Create temporary variables to help readability
  Population_t const * const pop_p = agent_p->pop_p;
  int * const grass_p = &(agent_p->cell_p->grass);

  // The prey should eat whichever is smaller: 1) The standard amount
  // it wants to eat in one sitting, 2) The amount of grass available in
  // the cell, or 3) The amount of grass it will take to make the prey
  // full
  int eatAmount = GetMin(pop_p->mealSize,
                         *grass_p,
                         pop_p->maxFood - agent_p->food);

  // Remove the grass from the cell
  (*grass_p) -= eatAmount;

  // Add the food to the prey
  agent_p->food += eatAmount;
}

// Define a function that returns the minimum of three numbers
int GetMin(int const a, int const b, int const c) {
  if (a <= b && a <= c) {
    return a;
  }
  else if (b <= a && b <= c) {
    return b;
  }
  else {
    return c;
  }
}

// Define a function that has a predator agent attempt to find and eat a prey
// agent
void APredatorTriesToEat(Agent_t * const agent_p) {
  // Create a temporary variable to help readability
  Population_t const * const pop_p = agent_p->pop_p;

  // Get a random neighbor cell with a prey in it
  Cell_t * neighborCell_p = GetRandomNeighborCell(agent_p->cell_p,
                                                  pop_p->preyType);

  // If no random prey was found, skip this predator
  if (neighborCell_p == NULL) {
    return;
  }

  // The predator should eat whichever is smaller: 1) The standard amount
  // it wants to eat in one sitting or 2) The amount of food it will take to
  // make the predator full
  agent_p->food = GetMin(pop_p->mealSize,
                         pop_p->mealSize,
                         pop_p->maxFood - agent_p->food);

  // Remove the prey
  RemoveAgent(neighborCell_p->agent_p);
}

// Define a function which, given a cell and an agent type, returns
// a pointer to random cell in one of the 4 neighboring directions
// that has the specified agent type. If no such cell is found, return NULL
Cell_t * GetRandomNeighborCell(Cell_t const * const cell_p,
                               AgentType_t const type) {
  Cell_t * neighborCell_ps[4];
  int neighborCellCount = 0;

  // Check if each of the cell's neighbor has an agent with the given type,
  // and, if so, adds it to the list
  for (int cellIdx = 0; cellIdx < cell_p->neighborCount; cellIdx++) {
    // Create temporary variables to help readability
    Cell_t * const neighbor_p = cell_p->neighbor_ps[cellIdx];
    Agent_t * const agent_p = neighbor_p->agent_p;

    // If the neighbor cell's agent has the same type we are looking for,
    // add the neighbor cell to the list
    if ((type == NONE && agent_p == NULL) ||
        (type != NONE && agent_p != NULL && agent_p->type == type)) {
      neighborCell_ps[neighborCellCount++] = neighbor_p;
    }
  }

  // If no suitable neighbors were found, return NULL
  if (neighborCellCount == 0) {
    return NULL;
  }

  // Otherwise, pick a random neighbor cell and return it
  return neighborCell_ps[GetRandom(neighborCellCount)];
}

// Define a function that, for all agents, creates a new agent of the same type 
// of a given agent at an a neighbor cell, provided the carrying capacity has
// not been reached, the percent chance of reproduction succeeds, and there is
// an open neighbor cell
void AllAgentsTryToReproduce() {
  // Loop over all the populations
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create temporary variables to help readability
    Population_t * const pop_p = &(Populations[popIdx]);
    Agent_t * const agents = pop_p->agents;

    // Make a copy of the count since we will be changing it during the loop
    // to add new agents
    int const count = pop_p->count;

    // Loop over all the agents in the population
    for (int agentIdx = 0; agentIdx < count; agentIdx++) {
      // If the carrying capacity has been reached, skip this population
      if (pop_p->count >= pop_p->capacity) {
        break;
      }

      // If the percent chance for reproduction fails, skip this agent
      if (GetRandom(100) >= pop_p->reprodChance) {
        continue;
      }

      // Create a temporary variable to help readability
      Agent_t * const agent_p = &(agents[agentIdx]);

      // Get a random open neighbor cell of the agent
      Cell_t * openNeighborCell_p = GetRandomNeighborCell(agent_p->cell_p,
                                                          NONE);

      // If there is no open neighbor cell, skip this agent
      if (openNeighborCell_p == NULL) {
        continue;
      }

      // Create a new agent at the cell
      AddAgent(pop_p, openNeighborCell_p, agent_p->food);
    }
  }
}

// Define a function which, for each agent, picks a random open neighbor cell
// and moves the agent into it
void AllAgentsTryToMove() {
  // Loop over each population
  for (int popIdx = 0; popIdx < POPULATION_COUNT; popIdx++) {
    // Create a temporary variable to help readability
    Population_t * const pop_p = &(Populations[popIdx]);

    // Loop over each agent
    for (int agentIdx = 0; agentIdx < pop_p->count; agentIdx++) {
      Agent_t * const agent_p = &(pop_p->agents[agentIdx]);

      // Get one of the agent's open neighbor cells
      Cell_t * const openNeighborCell_p = GetRandomNeighborCell(agent_p->cell_p,
                                                                NONE);

      // If there is no open neighbor, skip this agent
      if (openNeighborCell_p == NULL) {
        continue;
      } 

      // Move the agent
      openNeighborCell_p->agent_p = agent_p;
      agent_p->cell_p->agent_p = NULL;
      agent_p->cell_p = openNeighborCell_p;
    }
  }
}

// Define a function which, for each cell, grows grass if the grass amount is
// not at the max
void GrassGrows() {
  for (int y = 0; y < CELL_COUNT_Y; y++) {
    for (int x = 0; x < CELL_COUNT_X; x++) {
      // If this cell's grass amount is at max, skip this cell
      if (Cells[y][x].grass >= MAX_GRASS) {
        continue;
      }
      // Grow the grass
      Cells[y][x].grass++;
    }
  }
}

// Define a function that wraps up the model
void Finalize() {
  FreeMemory();
}

// Define a function that frees the allocated memory for the arrays of each
// population
void FreeMemory() {
  for (int popIdx = POPULATION_COUNT - 1; popIdx >= 0; popIdx--) {
    free(Populations[popIdx].agents);
  }
}

